#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <unistd.h>
#include <sys/wait.h>
#include <semaphore.h>

typedef struct {
  int pos;
  int cnt;
  int size;
  sem_t sem;
  char buffer[];
} pipe_t;

/* Opret en pipe */
pipe_t *pipe_create(unsigned int size) {
    pipe_t *p;
    int shmid;

    if (size < 1) {
      return NULL;
    }
    
    /* Opret et delt hukommelsessegment med strrelsen sizeof(pipe_t) + size */
    if ((shmid = shmget(IPC_PRIVATE, sizeof(pipe_t) + size, 0600 | IPC_CREAT)) == -1) {
      perror("shmget");
      exit(1);
    }

 /* Lg segmentet ind i processens adresserum. 'data' peger nu p
    den delte hukommelse */
    p = shmat(shmid, NULL, 0);
    if (p == (void *)(-1)) {
        perror("shmat");
        exit(1);
    }

    /* Initialiser semafor til 1. Dvs. vi kan g igang med at lave noget med det samme - bde parent og child */
    if (sem_init(&p->sem, 1, 1) == -1) {
        perror("sem_init");
        exit(1);
    }

    if (p != NULL) {
      p->pos = 0;
      p->cnt = 0;
      p->size = size;
    }
    
    return p;
}

/* Nedlg en pipe */
void pipe_free(pipe_t *p) {

  /* Udtag pipen fra adresserummet */
  if (shmdt(p) == -1) {
    perror("shmdt");
    exit(1);
  }
}

/*
Skriv 'len' bytes fra 'data' til pipe'en
Hvis der ikke er 'len' bytes fri, skrives
indtil pipe'en er fuld. Antallet af skrevne bytes
returneres. Funktionen er alts ikke-blokerende.
*/
int pipe_write(pipe_t *p, void *data, int len) {
  
  sem_wait(&p->sem);
  char *d = data;
  int i=0;

    while (p->cnt < p->size && i < len) {
        p->buffer[(p->pos + p->cnt) % p->size] = d[i];
        p->cnt++;
        i++;
    }
    sem_post(&p->sem);
    return i;
}

/*
Ls 'len' bytes fra pipe'en ind i 'data'.
Hvis der ikke er 'len' bytes tilgnglige lses
dem der er tilgnglige og det lste antal returneres.
Funktionen er alts ikke-blokerende.
*/
int pipe_read(pipe_t *p, void *data, int len) {
  sem_wait(&p->sem);
    char *d = data;
    int i=0;

    while (p->cnt > 0 && i < len) {
        d[i] = p->buffer[p->pos];
        p->cnt--;
        p->pos = (p->pos + 1) % p->size;
        i++;
    }
    sem_post(&p->sem);
    return i;
}

/* Lser noget data fra pipen og printer det til skrmen */
int print_pipe(pipe_t *p, int length, int pid) {
  int rlen;
  char str[length];
  
  rlen = pipe_read(p, str, length);
  str[rlen] = 0; /* terminer strengen */
  printf("pid: %d, len=%d: %s\n", pid, rlen, str);
  return rlen;
}

int main(int argc, const char* argv[]) {

  int size, rpos=0, wpos=0, slen, pid;
  pipe_t *p;

  char *test = "Jeg er havren, jeg har bjlder p... Bliv spejder min dreng!";
  slen = strlen(test);
  
  if (argc != 2) {
    printf("Usage: pipe <buffer size>\n");
    exit(1);
  }
  
  size = atoi(argv[1]);
  
  p = pipe_create(size);
  if (p == NULL) {
    printf("Error: could not create pipe!\n");
    exit(1);
  }

  /* Del i to processer */
  pid = fork();
  
  /* Child process */
  if (pid == 0) {
    while (wpos < slen) {
      wpos += pipe_write(p, test + wpos, slen - wpos);
    }
  }
  /* Parent process */
  else {
      while(rpos < slen) {
      rpos += print_pipe(p, rand()%size, pid);
    }
    wait(NULL); /* Vent p at child terminerer. */
  }

  pipe_free(p);
  
  return 0;
}
